@@ -10,6 +10,7 @@ from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives.asymmetric import ec
 from cryptography.hazmat.primitives.asymmetric import rsa
 
+from jwcrypto import jwa
 from jwcrypto import jwe
 from jwcrypto import jwk
 from jwcrypto import jws
@@ -170,6 +171,70 @@ RSAPrivateKey = {"kty": "RSA",
                        "eAvmj4sm-Fp0oYu_neotgQ0hzbI5gry7ajdYy9-2lNx_76aBZoOUu"
                        "9HCJ-UsfSOI8"}
 
+# From
+# vectors/cryptography_vectors/asymmetric/PEM_Serialization/rsa_private_key.pem
+RSAPrivatePEM = b"""-----BEGIN RSA PRIVATE KEY-----
+Proc-Type: 4,ENCRYPTED
+DEK-Info: DES-EDE3-CBC,B4B3C8C536E57CBE
+
+B8Lq1K/wcOr4JMspWrX3zCX14WAp3xgHsKAB4XfuCuju/HQZoWXtok1xoi5e2Ovw
+ENA99Jvb2yvBdDUfOlp1L1L+By3q+SwcdeNuEKjwGFG6MY2uZaVtLSiAFXf1N8PL
+id7FMRGPIxpTtXKMhfAq4luRb0BgKh7+ZvM7LkxkRxF7M1XVQPGhrU0OfxX9VODe
+YFH1q47os5JzHRcrRaFx6sn30e79ij2gRjzMVFuAX07n+yw3qeyNQNYdmDNP7iCZ
+x//0iN0NboTI81coNlxx7TL4bYwgESt1c2i/TCfLITjKgEny7MKqU1/jTrOJWu85
+PiK/ojaD1EMx9xxVgBCioQJVG/Jm9y+XhtGFAJUShzzsabX7KuANKRn3fgUN+yZS
+yp8hmD+R5gQHJk/8+zZ6/Imv8W/G+7fPZuSMgWeWtDReCkfzgnyIdjaIp3Pdp5yN
+WLLWADI4tHmNUqIzY7T25gVfg0P2tgQNzn3WzHxq4SfZN9Aw57woi8eSRpLBEn+C
+JjqwTxtFQ14ynG6GPsBaDcAduchmJPL7e9PuAfFyLJuM8sU8QyB2oir1M/qYFhTC
+ClXw2yylYjAy8TFw1L3UZA4hfAflINjYUY8pgAtTAjxeD/9PhiKSoMEX8Q/8Npti
+1Db5RpAClIEdB6nPywj6BzC+6El3dSGaCV0sTQ42LD+S3QH8VCwTB2AuKq7zyuD6
+wEQopcbIOGHSir875vYLmWLmqR9MCWZtKj/dWfTIQpBsPsI2ssZn/MptNqyEN9TW
+GfnWoTuzoziCS5YmEq7Mh98fwP9Krb0abo3fFvu6CY3dhvvoxPaXahyAxBfpKArB
+9nOf3gzHGReWNiFUtNZlvueYrC5CnblFzKaKB+81Imjw6RXM3QtuzbZ71zp+reL8
+JeiwE/mriwuGbxTJx5gwQX48zA5PJ342CCrl7jMeIos5KXmYkWoU5hEuGM3tK4Lx
+VAoGqcd/a4cWHuLWub8fbhFkIDcxFaMF8yQi0r2LOmvMOsv3RVpyfgJ07z5b9X1B
+w76CYkjGqgr0EdU40VTPtNhtHq7rrJSzGbapRsFUpvqgnkEwUSdbY6bRknLETmfo
+H3dPf2XQwXXPDMZTW54QsmQ9WjundqOFI2YsH6dCX/kmZK0IJVBpikL8SuM/ZJLK
+LcYJcrNGetENEKKl6hDwTTIsG1y3gx6y3wPzBkyJ2DtMx9dPoCqYhPHsIGc/td0/
+r4Ix9TWVLIl3MKq3z+/Hszd7jOnrkflfmKeA0DgJlqVJsuxP75pbdiKS/hCKRf8D
+AFJLvt6JSGBnz9ZZCB4KrjpHK/k+X7p8Y65uc/aX5BLu8vyRqFduhg98GVXJmD7k
+0ggXnqqFnies6SpnQ45cjfKSGDx/NjY0AwoGPH8n8CL6ZagU6K1utfHIMrqKkJME
+F6KcPHWrQkECojLdMoDInnRirdRb/FcAadWBSPrf+6Nln4ilbBJIi8W/yzeM/WFj
+UKKNjk4W26PGnNO6+TO5h1EpocDI4fx6UYIMmFjnyaLdLrSn1/SzuLL6I7pYZ0Um
+8qI4aWjP9RiUvGYJirfAUjL5Vp9w4+osf1sGiioe0GH/1WVuHeQ93A==
+-----END RSA PRIVATE KEY-----
+"""
+
+RSAPrivatePassword = b"123456"
+
+# From
+# vectors/cryptography_vectors/asymmetric/PEM_Serialization/rsa_public_key.pem
+RSAPublicPEM = b"""-----BEGIN PUBLIC KEY-----
+MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAnR4AZ+tgWYql+S3MaTQ6
+zeIO1fKzFIoau9Q0zGuv/1oCAewXwxeDSSxw+/Z3GL1NpuuS9CpbR5EQ3d71bD0v
+0G+Sf+mShSl0oljG7YqnNSPzKl+EQ3/KE+eEButcwas6KGof2BA4bFNCw/fPbuhk
+u/d8sIIEgdzBMiGRMdW33uci3rsdOenMZQA7uWsM/q/pu85YLAVOxq6wlUCzP4FM
+Tw/RKzayrPkn3Jfbqcy1aM2HDlFVx24vaN+RRbPSnVoQbo5EQYkUMXE8WmadSyHl
+pXGRnWsJSV9AdGyDrbU+6tcFwcIwnW22jb/OJy8swHdqKGkuR1kQ0XqokK1yGKFZ
+8wIDAQAB
+-----END PUBLIC KEY-----
+"""
+
+# From cryptography/vectors/cryptography_vectors/x509/v1_cert.pem
+PublicCert = b"""-----BEGIN CERTIFICATE-----
+MIIBWzCCAQYCARgwDQYJKoZIhvcNAQEEBQAwODELMAkGA1UEBhMCQVUxDDAKBgNV
+BAgTA1FMRDEbMBkGA1UEAxMSU1NMZWF5L3JzYSB0ZXN0IENBMB4XDTk1MDYxOTIz
+MzMxMloXDTk1MDcxNzIzMzMxMlowOjELMAkGA1UEBhMCQVUxDDAKBgNVBAgTA1FM
+RDEdMBsGA1UEAxMUU1NMZWF5L3JzYSB0ZXN0IGNlcnQwXDANBgkqhkiG9w0BAQEF
+AANLADBIAkEAqtt6qS5GTxVxGZYWa0/4u+IwHf7p2LNZbcPBp9/OfIcYAXBQn8hO
+/Re1uwLKXdCjIoaGs4DLdG88rkzfyK5dPQIDAQABMAwGCCqGSIb3DQIFBQADQQAE
+Wc7EcF8po2/ZO6kNCwK/ICH6DobgLekA5lSLr5EvuioZniZp5lFzAw4+YzPQ7XKJ
+zl9HYIMxATFyqSiD9jsx
+-----END CERTIFICATE-----
+"""
+
+PublicCertThumbprint = u'7KITkGJF74IZ9NKVvHfuJILbuIZny6j-roaNjB1vgiA'
+
 
 class TestJWK(unittest.TestCase):
     def test_create_pubKeys(self):
@@ -198,8 +263,12 @@ class TestJWK(unittest.TestCase):
         jwk.JWK.generate(kty='oct', size=256)
         jwk.JWK.generate(kty='RSA', size=4096)
         jwk.JWK.generate(kty='EC', curve='P-521')
-        k = jwk.JWK.generate(kty='oct', size=256, kid='MySymmetricKey')
+        k = jwk.JWK.generate(kty='oct', alg='A192KW', kid='MySymmetricKey')
         self.assertEqual(k.key_id, 'MySymmetricKey')
+        self.assertEqual(len(base64url_decode(k.get_op_key('encrypt'))), 24)
+        jwk.JWK.generate(kty='RSA', alg='RS256')
+        k = jwk.JWK.generate(kty='RSA', size=4096, alg='RS256')
+        self.assertEqual(k.get_op_key('encrypt').key_size, 4096)
 
     def test_export_public_keys(self):
         k = jwk.JWK(**RSAPrivateKey)
@@ -267,6 +336,48 @@ class TestJWK(unittest.TestCase):
                 k.thumbprint(),
                 PublicKeys['thumbprints'][i])
 
+    def test_import_from_pem(self):
+        pubk = jwk.JWK.from_pem(RSAPublicPEM)
+        self.assertEqual(pubk.export_to_pem(), RSAPublicPEM)
+        rsapub = pubk.get_op_key('verify')
+
+        prik = jwk.JWK.from_pem(RSAPrivatePEM, password=RSAPrivatePassword)
+        rsapri = prik.get_op_key('sign')
+        self.assertEqual(rsapri.public_key().public_numbers().n,
+                         rsapub.public_numbers().n)
+
+        pubc = jwk.JWK.from_pem(PublicCert)
+        self.assertEqual(pubc.key_id, PublicCertThumbprint)
+
+    def test_export_symmetric(self):
+        key = jwk.JWK(**SymmetricKeys['keys'][0])
+        self.assertTrue(key.is_symmetric)
+        self.assertFalse(key.has_public)
+        self.assertFalse(key.has_private)
+        self.assertEqual(json_encode(SymmetricKeys['keys'][0]),
+                         key.export_symmetric())
+
+    def test_export_public(self):
+        key = jwk.JWK.from_pem(PublicCert)
+        self.assertFalse(key.is_symmetric)
+        self.assertTrue(key.has_public)
+        self.assertFalse(key.has_private)
+        pubc = key.export_public()
+        self.assertEqual(json_decode(pubc)["kid"], PublicCertThumbprint)
+
+    def test_export_private(self):
+        key = jwk.JWK.from_pem(RSAPrivatePEM, password=RSAPrivatePassword)
+        self.assertFalse(key.is_symmetric)
+        self.assertTrue(key.has_public)
+        self.assertTrue(key.has_private)
+        pri = key.export_private()
+        prikey = jwk.JWK(**json_decode(pri))
+        self.assertTrue(prikey.has_private)
+        pub = key.export_public()
+        pubkey = jwk.JWK(**json_decode(pub))
+        self.assertFalse(pubkey.has_private)
+        self.assertEqual(prikey.key_id, pubkey.key_id)
+
 
 # RFC 7515 - A.1
 A1_protected = \
@@ -503,7 +614,8 @@ class TestJWS(unittest.TestCase):
                           self.check_sign, A5_example)
         a5_bis = {'allowed_algs': ['none']}
         a5_bis.update(A5_example)
-        self.check_sign(a5_bis)
+        with self.assertRaises(jws.InvalidJWSSignature):
+            self.check_sign(a5_bis)
 
     def test_A6(self):
         s = jws.JWS(A6_example['payload'])
@@ -732,6 +844,87 @@ class TestJWE(unittest.TestCase):
             e.deserialize(E_A5_ex, E_A4_ex['key2'])
 
 
+MMA_vector_key = jwk.JWK(**E_A2_key)
+MMA_vector_ok_cek =  \
+    '{"protected":"eyJlbmMiOiJBMTI4Q0JDLUhTMjU2In0",' \
+    '"unprotected":{"jku":"https://server.example.com/keys.jwks"},' \
+    '"recipients":[' \
+    '{"header":{"alg":"RSA1_5","kid":"2011-04-29"},' \
+    '"encrypted_key":'\
+    '"UGhIOguC7IuEvf_NPVaXsGMoLOmwvc1GyqlIKOK1nN94nHPoltGRhWhw7Zx0-' \
+    'kFm1NJn8LE9XShH59_i8J0PH5ZZyNfGy2xGdULU7sHNF6Gp2vPLgNZ__deLKx' \
+    'GHZ7PcHALUzoOegEI-8E66jX2E4zyJKx-YxzZIItRzC5hlRirb6Y5Cl_p-ko3' \
+    'YvkkysZIFNPccxRU7qve1WYPxqbb2Yw8kZqa2rMWI5ng8OtvzlV7elprCbuPh' \
+    'cCdZ6XDP0_F8rkXds2vE4X-ncOIM8hAYHHi29NX0mcKiRaD0-D-ljQTP-cFPg' \
+    'wCp6X-nZZd9OHBv-B3oWh2TbqmScqXMR4gp_A"}],' \
+    '"iv":"AxY8DCtDaGlsbGljb3RoZQ",' \
+    '"ciphertext":"PURPOSEFULLYBROKENYGS4HffxPSUrfmqCHXaI9wOGY",' \
+    '"tag":"Mz-VPPyU4RlcuYv1IwIvzw"}'
+MMA_vector_ko_cek = \
+    '{"protected":"eyJlbmMiOiJBMTI4Q0JDLUhTMjU2In0",' \
+    '"unprotected":{"jku":"https://server.example.com/keys.jwks"},' \
+    '"recipients":[' \
+    '{"header":{"alg":"RSA1_5","kid":"2011-04-29"},' \
+    '"encrypted_key":'\
+    '"UGhIOguC7IuEvf_NPVaYsGMoLOmwvc1GyqlIKOK1nN94nHPoltGRhWhw7Zx0-' \
+    'kFm1NJn8LE9XShH59_i8J0PH5ZZyNfGy2xGdULU7sHNF6Gp2vPLgNZ__deLKx' \
+    'GHZ7PcHALUzoOegEI-8E66jX2E4zyJKx-YxzZIItRzC5hlRirb6Y5Cl_p-ko3' \
+    'YvkkysZIFNPccxRU7qve1WYPxqbb2Yw8kZqa2rMWI5ng8OtvzlV7elprCbuPh' \
+    'cCdZ6XDP0_F8rkXds2vE4X-ncOIM8hAYHHi29NX0mcKiRaD0-D-ljQTP-cFPg' \
+    'wCp6X-nZZd9OHBv-B3oWh2TbqmScqXMR4gp_A"}],' \
+    '"iv":"AxY8DCtDaGlsbGljb3RoZQ",' \
+    '"ciphertext":"PURPOSEFULLYBROKENYGS4HffxPSUrfmqCHXaI9wOGY",' \
+    '"tag":"Mz-VPPyU4RlcuYv1IwIvzw"}'
+
+
+class TestMMA(unittest.TestCase):
+    @classmethod
+    def setUpClass(cls):
+        import os
+        cls.enableMMA = os.environ.get('JWCRYPTO_TESTS_ENABLE_MMA', False)
+        cls.iterations = 500
+        cls.sub_iterations = 100
+
+    def test_MMA(self):
+        if self.enableMMA:
+
+            print('Testing MMA timing attacks')
+
+            ok_cek = 0
+            ok_e = jwe.JWE()
+            ok_e.deserialize(MMA_vector_ok_cek)
+            ko_cek = 0
+            ko_e = jwe.JWE()
+            ko_e.deserialize(MMA_vector_ko_cek)
+
+            import time
+            counter = getattr(time, 'perf_counter', time.time)
+
+            for _ in range(self.iterations):
+                start = counter()
+                for _ in range(self.sub_iterations):
+                    with self.assertRaises(jwe.InvalidJWEData):
+                        ok_e.decrypt(MMA_vector_key)
+                stop = counter()
+                ok_cek += (stop - start) / self.sub_iterations
+
+                start = counter()
+                for _ in range(self.sub_iterations):
+                    with self.assertRaises(jwe.InvalidJWEData):
+                        ko_e.decrypt(MMA_vector_key)
+                stop = counter()
+                ko_cek += (stop - start) / self.sub_iterations
+
+            ok_cek /= self.iterations
+            ko_cek /= self.iterations
+
+            deviation = ((ok_cek - ko_cek) / ok_cek) * 100
+            print('MMA ok cek: {}'.format(ok_cek))
+            print('MMA ko cek: {}'.format(ko_cek))
+            print('MMA deviation: {}% ({})'.format(int(deviation), deviation))
+            self.assertLess(deviation, 2)
+
+
 # RFC 7519
 A1_header = {
     "alg": "RSA1_5",
@@ -883,3 +1076,75 @@ class ConformanceTests(unittest.TestCase):
         check.deserialize(o, jwk.JWK(kty='oct', k=base64url_encode(b'A' * 16)),
                           alg="HS512")
         self.assertTrue(check.objects['valid'])
+
+    def test_jws_headers_as_dicts(self):
+        sign = jws.JWS(payload='message')
+        key = jwk.JWK(kty='oct', k=base64url_encode(b'A' * 16))
+        sign.add_signature(key, protected={'alg': 'HS512'},
+                           header={'kid': key.thumbprint()})
+        o = sign.serialize()
+        check = jws.JWS()
+        check.deserialize(o, key, alg="HS512")
+        self.assertTrue(check.objects['valid'])
+        self.assertEqual(check.jose_header['kid'], key.thumbprint())
+
+    def test_jwe_headers_as_dicts(self):
+        enc = jwe.JWE(plaintext='message',
+                      protected={"alg": "A256KW", "enc": "A256CBC-HS512"})
+        key = jwk.JWK(kty='oct', k=base64url_encode(b'A' * 32))
+        enc.add_recipient(key, {'kid': key.thumbprint()})
+        o = enc.serialize()
+        check = jwe.JWE()
+        check.deserialize(o)
+        check.decrypt(key)
+        self.assertEqual(check.payload, b'message')
+        self.assertEqual(
+            json_decode(check.objects['header'])['kid'], key.thumbprint())
+
+    def test_jwe_default_recipient(self):
+        key = jwk.JWK(kty='oct', k=base64url_encode(b'A' * (128 // 8)))
+        enc = jwe.JWE(plaintext='plain',
+                      protected='{"alg":"A128KW","enc":"A128GCM"}',
+                      recipient=key).serialize()
+        check = jwe.JWE()
+        check.deserialize(enc, key)
+        self.assertEqual(b'plain', check.payload)
+
+
+class JWATests(unittest.TestCase):
+    def test_jwa_create(self):
+        for name, cls in jwa.JWA.algorithms_registry.items():
+            self.assertEqual(cls.name, name)
+            self.assertIn(cls.algorithm_usage_location, {'alg', 'enc'})
+            if name == 'ECDH-ES':
+                self.assertIs(cls.keysize, None)
+            else:
+                self.assertIsInstance(cls.keysize, int)
+                self.assertGreaterEqual(cls.keysize, 0)
+
+            if cls.algorithm_use == 'sig':
+                with self.assertRaises(jwa.InvalidJWAAlgorithm):
+                    jwa.JWA.encryption_alg(name)
+                with self.assertRaises(jwa.InvalidJWAAlgorithm):
+                    jwa.JWA.keymgmt_alg(name)
+                inst = jwa.JWA.signing_alg(name)
+                self.assertIsInstance(inst, jwa.JWAAlgorithm)
+                self.assertEqual(inst.name, name)
+            elif cls.algorithm_use == 'kex':
+                with self.assertRaises(jwa.InvalidJWAAlgorithm):
+                    jwa.JWA.encryption_alg(name)
+                with self.assertRaises(jwa.InvalidJWAAlgorithm):
+                    jwa.JWA.signing_alg(name)
+                inst = jwa.JWA.keymgmt_alg(name)
+                self.assertIsInstance(inst, jwa.JWAAlgorithm)
+                self.assertEqual(inst.name, name)
+            elif cls.algorithm_use == 'enc':
+                with self.assertRaises(jwa.InvalidJWAAlgorithm):
+                    jwa.JWA.signing_alg(name)
+                with self.assertRaises(jwa.InvalidJWAAlgorithm):
+                    jwa.JWA.keymgmt_alg(name)
+                inst = jwa.JWA.encryption_alg(name)
+                self.assertIsInstance(inst, jwa.JWAAlgorithm)
+                self.assertEqual(inst.name, name)
+            else:
+                self.fail((name, cls))
